"""Utils."""
import functools
from typing import Callable
from datetime import datetime, timedelta


class ExpirableCache(object):
    """aiosonic.utils.ExpirableCache class.

    Class used for custom cache decorator, dummy expirable cache based
    on dict structure.

    Params:
        * **size**: max items in dict.
        * **timeout**: Timeout in milliseconds, if it is None,
            there is no timeout.
    """

    def __init__(self, size=512, timeout=None):
        self.cache = {}
        self.timeout = timeout
        self.size = size

    def set(self, key, data):
        if self.timeout:
            expire_at = datetime.utcnow() + timedelta(
                milliseconds=self.timeout)
            self.cache[key] = {
                'value': data,
                'expire_at': expire_at
            }
        else:
            self.cache[key] = data

        if len(self.cache) > self.size:
            self.cache.pop(next(iter(self.cache)))

    def get(self, key):
        data = self.cache.get(key)
        if self.timeout and data:
            if datetime.utcnow() > data['expire_at']:
                del self.cache[key]
                data = None
            return data['value']
        return data

    def __len__(self):
        return len(self.cache)


def cache_decorator(size: int=512, timeout: int = None) -> Callable:
    """Dummy cache decorator."""
    cache = ExpirableCache(size, timeout)

    def decorator(func):
        def _wrapper(*args):
            key = '-'.join(list(args))
            data = cache.get(key)
            if data:
                return data

            res = func(key)

            cache.set(key, data)
            return res

        _wrapper.cache = cache
        return _wrapper

    return decorator



"""Connection stuffs."""

import ssl
from ssl import SSLContext
from asyncio import open_connection
from asyncio import StreamReader
from asyncio import StreamWriter
from typing import Dict
from typing import Optional
from urllib.parse import ParseResult

import h2.connection
import h2.events

# from concurrent import futures (unused)
from aiosonic.exceptions import HttpParsingError
from aiosonic.timeout import Timeouts
from aiosonic.connectors import TCPConnector
from aiosonic.http2 import Http2Handler

from aiosonic.types import ParsedBodyType


class Connection:
    """Connection class."""

    def __init__(self, connector: TCPConnector) -> None:
        self.connector = connector
        self.reader: Optional[StreamReader] = None
        self.writer: Optional[StreamWriter] = None
        self.keep = False  # keep alive flag
        self.key = None
        self.blocked = False
        self.temp_key: Optional[str] = None

        self.h2conn: Optional[h2.connection.H2Connection] = None
        self.h2handler: Optional[Http2Handler] = None

    async def connect(self,
                      urlparsed: ParseResult,
                      dns_info: dict,
                      verify: bool,
                      ssl_context: SSLContext,
                      http2: bool = False) -> None:
        """Connet with timeout."""
        await self._connect(
            urlparsed, verify, ssl_context, dns_info, http2
        )

    async def _connect(self, urlparsed: ParseResult, verify: bool,
                       ssl_context: SSLContext, dns_info, http2: bool) -> None:
        """Get reader and writer."""
        if not urlparsed.hostname:
            raise HttpParsingError('missing hostname')

        key = f'{urlparsed.hostname}-{urlparsed.port}'

        if self.writer:
            # python 3.6 doesn't have writer.is_closing
            is_closing = getattr(
                self.writer, 'is_closing',
                self.writer._transport.is_closing)  # type: ignore
        else:

            def is_closing():
                return True  # noqa

        dns_info_copy = dns_info.copy()
        dns_info_copy['server_hostname'] = dns_info_copy.pop('hostname')

        if not (self.key and key == self.key and not is_closing()):
            self.close()

            if urlparsed.scheme == 'https':
                ssl_context = ssl_context or ssl.create_default_context(
                    ssl.Purpose.SERVER_AUTH, )
                # flag will be removed when fully http2 support
                if http2:  # pragma: no cover
                    ssl_context.set_alpn_protocols(['h2', 'http/1.1'])
                if not verify:
                    ssl_context.check_hostname = False
                    ssl_context.verify_mode = ssl.CERT_NONE
            else:
                del dns_info_copy['server_hostname']
            port = urlparsed.port or (443
                                      if urlparsed.scheme == 'https' else 80)
            dns_info_copy['port'] = port
            self.reader, self.writer = await open_connection(
                **dns_info_copy, ssl=ssl_context)

            self.temp_key = key
            await self._connection_made()

    async def _connection_made(self) -> None:
        tls_conn = self.writer.get_extra_info('ssl_object')
        if not tls_conn:
            return

        # Always prefer the result from ALPN to that from NPN.
        negotiated_protocol = tls_conn.selected_alpn_protocol()
        if negotiated_protocol is None:
            negotiated_protocol = tls_conn.selected_npn_protocol()

        if negotiated_protocol == 'h2':  # pragma: no cover
            self.h2conn = h2.connection.H2Connection()
            self.h2handler = Http2Handler(self)

    def keep_alive(self) -> None:
        """Check if keep alive."""
        self.keep = True

    def block_until_read_chunks(self):
        """Check if keep alive."""
        self.blocked = True

    async def __aenter__(self):
        """Get connection from pool."""
        return self

    async def __aexit__(self, exc_type: None, exc: None, tb: None) -> None:
        """Release connection."""
        if self.keep and not exc:
            self.key = self.temp_key
        else:
            self.key = None
            self.h2conn = None
            if self.writer and not self.blocked:
                self.close()

        if not self.blocked:
            await self.release()
            if self.h2handler:  # pragma: no cover
                self.h2handler.cleanup()

    async def release(self) -> None:
        """Release connection."""
        await self.connector.release(self)
        # if keep False and blocked (by latest chunked response), close it.
        # server said to close it.
        if not self.keep and self.blocked:
            self.blocked = False
            self.close()
        # ensure unblock conn object after read
        self.blocked = False

    def __del__(self) -> None:
        """Cleanup."""
        self.close(True)

    def close(self, check_closing: bool = False) -> None:
        """Close connection if opened."""
        if self.writer:
            is_closing = getattr(self.writer, 'is_closing',
                                 self.writer._transport.is_closing)
            if not check_closing or is_closing():
                self.writer.close()

    async def http2_request(self, headers: Dict[str, str],
                            body: Optional[ParsedBodyType]):
        if self.h2handler:  # pragma: no cover
            return await self.h2handler.request(headers, body)



"""Connector stuffs."""
import random
from asyncio import sleep as asyncio_sleep
from asyncio import wait_for
from ssl import SSLContext
from typing import Coroutine
from urllib.parse import ParseResult

# import h2.connection (unused)
from hyperframe.frame import SettingsFrame

# from concurrent import futures (unused)
from aiosonic.exceptions import (ConnectionPoolAcquireTimeout, ConnectTimeout,
                                 HttpParsingError, TimeoutException)
from aiosonic.pools import SmartPool
from aiosonic.resolver import DefaultResolver
from aiosonic.timeout import Timeouts
from aiosonic.utils import ExpirableCache


class TCPConnector:
    """TCPConnector.

    Holds the main logic for making connections to destination hosts.

    Params:
        * **pool_size**: size for pool of connections
        * **timeouts**: global timeouts to use for connections with this connector. default: :class:`aiosonic.timeout.Timeouts` instance with default args.
        * **connection_cls**: connection class to be used. default: :class:`aiosonic.connection.Connection`
        * **pool_cls**: pool class to be used. default: :class:`aiosonic.pools.SmartPool`
        * **resolver**: resolver to be used. default: :class:`aiosonic.resolver.DefaultResolver`
        * **ttl_dns_cache**: ttl in milliseconds for dns cache. default: `10000` 10 seconds
        * **use_dns_cache**: Flag to indicate usage of dns cache. default: `True`

    """

    def __init__(self,
                 pool_size: int = 25,
                 timeouts: Timeouts = None,
                 connection_cls=None,
                 pool_cls=None,
                 resolver=None,
                 ttl_dns_cache=10000,
                 use_dns_cache=True):
        from aiosonic.connection import Connection  # avoid circular dependency
        self.pool_size = pool_size
        connection_cls = connection_cls or Connection
        pool_cls = pool_cls or SmartPool
        self.pool = pool_cls(self, pool_size, connection_cls)
        self.timeouts = timeouts or Timeouts()
        self.resolver = resolver or DefaultResolver()
        self.use_dns_cache = use_dns_cache
        if self.use_dns_cache:
            self.cache = ExpirableCache(512, ttl_dns_cache)

    async def acquire(self, urlparsed: ParseResult, verify, ssl, timeouts, http2):
        """Acquire connection."""
        if not urlparsed.hostname:
            raise HttpParsingError('missing hostname')

        # Faster without timeout
        if not self.timeouts.pool_acquire:
            conn = await self.pool.acquire(urlparsed)
            return await self.after_acquire(
                urlparsed, conn, verify, ssl, timeouts, http2)

        try:
            conn = await wait_for(self.pool.acquire(urlparsed),
                                  self.timeouts.pool_acquire)
            return await self.after_acquire(
                urlparsed, conn, verify, ssl, timeouts, http2)
        except TimeoutException:
            raise ConnectionPoolAcquireTimeout()

    async def after_acquire(self, urlparsed, conn, verify, ssl, timeouts, http2):
        dns_info = await self.__resolve_dns(
            urlparsed.hostname, urlparsed.port)

        try:
            await wait_for(conn.connect(
                urlparsed, dns_info, verify, ssl, http2
            ), timeout=timeouts.sock_connect)
        except TimeoutException:
            raise ConnectTimeout()
        return conn

    async def release(self, conn):
        """Release connection."""
        res = self.pool.release(conn)
        if isinstance(res, Coroutine):
            await res

    async def wait_free_pool(self):
        """Wait until free pool."""
        while True:
            if self.pool.is_all_free():
                return True
            asyncio_sleep(0.02)  # pragma: no cover

    async def cleanup(self):
        """Cleanup connector connections."""
        await self.pool.cleanup()

    async def __resolve_dns(self, host: str, port: int):
        key = f'{host}-{port}'
        dns_data = self.cache.get(key)
        if not dns_data:
            dns_data = await self.resolver.resolve(host, port)
            self.cache.set(key, dns_data)
        return random.choice(dns_data)
